(*pp deriving *)
type comparison = [`Less | `LessEq | `Equal | `NotEq]
    deriving (Eq, Show, Pickle, Typeable, Shelve)

type label = string
    deriving (Pickle)

type constant =
  | Boolean of bool
  | Integer of Num.num
  | Char of char
  | String of string
  | Float of float
      deriving (Eq)

module type PreExpr = sig
  type 'a datatype'
  type query

  type 'a expression' =
  | Constant of (constant * 'a)
  | Variable of (string * 'a)
  | Abs of ('a expression' * 'a)
  | App of ('a expression' * 'a expression' * 'a)
  | Apply of ('a expression' * 'a expression' list * 'a)
  | Condition of ('a expression' * 'a expression' * 'a expression' * 'a)
  | Comparison of ('a expression' * comparison * 'a expression' * 'a)
  | Abstr of ('a expression' list * 'a expression' * 'a)
  | Let of (string * 'a expression' * 'a expression' * 'a)
  | Rec of ((string * 'a expression' * 'a datatype' option) list * 'a expression' * 'a)
  | Xml_node of (string * (string * 'a expression') list * 'a expression' list * 'a)
  | Record_intro of (('a expression') Utility.stringmap * ('a expression') option * 'a)
  | Record_selection of (string * string * string * 'a expression' * 'a expression' * 'a)
  | Record_unpack of ((string * string) list * 'a expression' * 'a expression' * 'a expression' option * 'a)
  | Record_pack of (string * 'a expression' * 'a datatype' * 'a expression' * 'a)
  | Project of ('a expression' * string * 'a)
  | Erase of ('a expression' * string * 'a)
  | Variant_injection of (string * 'a expression' * 'a)
  | Variant_selection of ('a expression' * string * string * 'a expression' *
                            string * 'a expression' * 'a)
  | Variant_selection_empty of ('a expression' * 'a)
  | Nil of 'a
  | List_of of ('a expression' * 'a)
  | Concat of ('a expression' * 'a expression' * 'a)
  | For of ('a expression' * string * 'a expression' * 'a)
  | Database of ('a expression' * 'a)
  | TableQuery of ((string * 'a expression') list * query * 'a)
      (** (table_alias * Variable (..)) list * query * 'a  *)
  | TableHandle of ('a expression' * 'a expression' * 'a table_constr * 'a)
  | ViewQuery of ('a expression' * 'a)
  | ViewHandle of ('a expression' * 'a expression' * 'a expression' * 'a)
  | SortBy of ('a expression' * 'a expression' * 'a)
  | Call_cc of ('a expression' * 'a)
  | Wrong of 'a
  | HasType of ('a expression' * 'a datatype' * 'a)   
  and 'a table_constr = {readtype:'a datatype'; writetype: 'a datatype'; marshallby: (string * 'a expression') list; primarykey: string list}
end
  
type position = Lexing.position * string * string
and untyped_data = [`U of position]
and 't annotated_data = [`T of (position * 't * label option)]
    deriving(Eq, Show, Typeable, Pickle, Shelve)    

module rec Expr : (PreExpr with type 'a datatype' = 'a Types.datatype'
			      and type query = Query.query)
and Query : (Mkquery.Q with type datatype = Types.datatype)
and Types : (
  Mktypes.T with type 'a fq = 'a FableType.fable_qualifier
		    and type kind = FableType.kind
		    and type typed_data = FableType.typed_data
		    and type untyped_data = FableType.untyped_data)
and FableType : (
  Mktypes.FableType with type 'a expression' = 'a Expr.expression'
				    and type 'a datatype' = 'a Types.datatype'
				    and type untyped_data = untyped_data
				    and type typed_data = Types.datatype annotated_data)
and DerivingSyntax : sig
  open Eq
  open Show
  open Pickle
  open Typeable
  open Shelve

  module Eq_datatype : Eq with type a = Types.datatype
  module Show_datatype : Show with type a = Types.datatype
  module Pickle_datatype : Pickle with type a = Types.datatype
  module Typeable_datatype : Typeable with type a = Types.datatype
  module Shelve_datatype : Shelve with type a = Types.datatype

  module Show_ldatatype : Show with type a = Types.ldatatype

  module Eq_row : Eq with type a = Types.row
  module Show_row : Show with type a = Types.row
  module Pickle_row : Pickle with type a = Types.row
  module Typeable_row : Typeable with type a = Types.row
  module Shelve_row : Shelve with type a = Types.row

  module Eq_typed_expression : Eq with type a = FableType.typed_expression
  module Show_typed_expression : Show with type a = FableType.typed_expression
  module Show_debug_expression : Show with type a = FableType.typed_expression
  module Pickle_typed_expression : Pickle with type a = FableType.typed_expression
  module Typeable_typed_expression : Typeable with type a = FableType.typed_expression
  module Shelve_typed_expression : Shelve with type a = FableType.typed_expression
   
  module Eq_u_datatype : Eq with type a = Types.u_datatype
  module Show_u_datatype : Show with type a = Types.u_datatype
  module Eq_untyped_expression : Eq with type a = FableType.untyped_expression
  module Show_untyped_expression : Show with type a = FableType.untyped_expression

  module Show_stripped_expression : Show with type a = unit FableType.expression'
  module Show_assumption : Show with type a = Types.assumption
  module Show_u_assumption : Show with type a = Types.u_assumption
  module Show_fable_type : Show with type a = FableType.fable_type
    
  module Show_table_constr : Show with type a = FableType.typed_data Expr.table_constr

  module Functor_expression' : sig
    val map : ('a -> 'b) -> 'a Expr.expression' -> 'b Expr.expression'
  end
  
  module Rewrite_expression' (T : sig type a end) : 
    Rewrite.Rewrite with type t = T.a Expr.expression'
end

