(*pp deriving *)
open Syntax
open Expression
open Expression.Types

type omega
type environment = (string * Types.assumption) list
type alias_environment = Types.assumption Mktypes.stringmap
type typing_environment = environment * alias_environment

exception Shadow_variable of string
exception UndefinedAlias of string
exception AliasMismatch of string

val empty : omega
val initial_env : typing_environment -> omega

val check_context : omega -> Syntax.location -> bool
val get_context : omega -> Syntax.location 
val set_context : omega -> Syntax.location -> omega
val add_rec_var : int -> omega -> omega
val contains_rec_var : int -> omega -> bool

val add_refinement : (Syntax.expression * Syntax.expression) -> omega -> omega
val refine : (Syntax.expression -> Syntax.expression -> bool) -> Syntax.expression -> omega -> Syntax.expression option
val get_type_annotation : omega -> datatype option
val set_type_annotation : omega -> datatype -> omega
val set_type_annotation_opt : omega -> datatype option -> omega
val clear_type_annotation : omega -> omega

val set_phantom_phase : omega -> omega
val clear_phantom_phase : omega -> omega
  
val lookup : string -> omega -> Types.assumption
val lookup_location : string -> omega -> location
val contains : string -> omega -> bool

val register_alias : (string * (int * FableType.kind) list * Types.datatype) -> omega -> omega
val free_alias_check : omega -> datatype -> unit
val free_alias_check_row : omega -> row -> unit

val is_mailbox_free : omega -> datatype -> bool
val is_mailbox_free_row : omega -> row -> bool
  
val add_phantom_binding : string -> Types.assumption -> omega -> omega
val add_noshadow_binding : string -> Types.assumption -> omega -> omega
val add_normal_binding : ?location:location -> string -> Types.assumption -> omega -> omega 
  
val alias_env : omega -> alias_environment
val lookup_alias : (string * Types.datatype list) -> omega -> Types.assumption
val concat_environment : omega -> omega -> omega
val all_bound_names : omega -> Utility.StringSet.t
val all_type_vars : omega -> Types.TypeVarSet.t

(* Deprecated *)
val extract_typing_environment : omega -> typing_environment
val lookup_alias_old : (string * Types.datatype list) -> alias_environment -> Types.assumption
